//
// Programmer:    Craig Stuart Sapp <craig@ccrma.stanford.edu>
// Thanks to:     Erik Neuenschwander <erikn@leland.stanford.edu>
//                for Windows 95 assembly code for timing
// Creation Date: Mon Oct 13 11:34:57 GMT-0800 1997
// Last Modified: Tue Feb 10 21:05:28 GMT-0800 1998
// Filename:      .../sig/code/control/Timer/Timer.cpp
// Syntax:        C++ 
// $Smake:        cc -g -c %b.cpp -I../../../include && rm -f %b.o
//


#include "Timer.h"
#include <iostream.h>
#include <stdlib.h>


#ifdef VISUAL
   #define WIN32_LEAN_AND_MEAN
   #include <windows.h>
   #undef WIN32_LEAN_AND_MEAN
#else
   #include <unistd.h>
   #include <libc.h>
#endif


// declare static variables
int64bits Timer::globalOffset = 0;
long      Timer::cpuSpeed     = 200000000;  // 200 MHz



//////////////////////////////
//
// Timer::Timer
//

Timer::Timer(void) {
   if (globalOffset == 0) {             // coordinate offset between timers
      globalOffset = clockCycles();
   }
   offset = globalOffset;
   ticksPerSecond = 1000;
   period = 1000.0;                     // default period of once per second
   cpuSpeed = measureCpuSpeed();
}


Timer::Timer(long aSpeed) {
   if (globalOffset == 0) {
      globalOffset = clockCycles();
   }
   offset = globalOffset;
   cpuSpeed = aSpeed;
   ticksPerSecond = 1000;
   period = 1000.0;                     // default period of once per second
}


Timer::Timer(Timer& aTimer) {
   offset = aTimer.offset;
   cpuSpeed = aTimer.cpuSpeed;
   ticksPerSecond = aTimer.ticksPerSecond;
   period = aTimer.period;
}



//////////////////////////////
//
// Timer::~Timer
//

Timer::~Timer() {
   // do nothing
}



//////////////////////////////
//
// Timer::clockCycles -- returns the number of clock cycles since last reboot
//	HARDWARE DEPENDENT -- currently for Pentiums only.
//


int64bits Timer::clockCycles() const {
#ifndef VISUAL
   int64bits output;

   // For Pentiums, you can get the number of clock cycles elapsed 
   // since the last reboot with the following assembler code:
   __asm__ volatile (".byte 0x0f, 0x31" : "=A" (output));

#else
   int64bits output;
   unsigned long high_end, low_end;
   __asm {
      __asm _emit 0x0f __asm _emit 0x31
      mov high_end, edx
      mov low_end, eax
   }
   output = high_end;
   output = output << 32;
   output += low_end;
#endif

   return output;
}



//////////////////////////////
//
// Timer::expired -- returns the number of periods
//    That have passed since the last update or reset.
//


double Timer::expired(void) const {
   return (double)getTime()/period;
}



//////////////////////////////
//
// Timer::getCpuSpeed -- returns the CPU speed of the computer.
//

long Timer::getCpuSpeed(void) const {
   return cpuSpeed;
}



//////////////////////////////
//
// Timer::getFactor 
//

double Timer::getFactor(void) const{
   return (double)((double)getCpuSpeed()/(double)getTicksPerSecond());
}



//////////////////////////////
//
// Timer::getPeriod -- returns the timing period of the timer,
//    if the timer is being used as a periodic timer.
//

double Timer::getPeriod(void) const {
   return period;
}



//////////////////////////////
//
// Timer::getTicksPerSecond
//

long Timer::getTicksPerSecond(void) const {
   return ticksPerSecond;
}



//////////////////////////////
//
// Timer::getTime -- returns time in milliseconds by default.  
// 	time can be altered to be in some other unit of time
//	by using the setTicksPerSecond function.
//	(Default is 1000 ticks per second.)
//

long Timer::getTime(void) const {
   return (long)((clockCycles()-offset)/getFactor());
}
   


//////////////////////////////
//
// Timer::getTimeInSeconds 
//

double Timer::getTimeInSeconds(void) const {
   return ((clockCycles()-offset)/(double)cpuSpeed);
}
   


//////////////////////////////
//
// Timer::getTimeInTicks 
//

long Timer::getTimeInTicks(void) const {
   return (long)((clockCycles()-offset)/getFactor());
}
   


//////////////////////////////
//
// Timer::measureCpuSpeed -- returns the number of clock cycles in 
//	one second.  Accuracy to about +/- 5%.
//

long Timer::measureCpuSpeed(void) const {
   int64bits a, b;
   a = clockCycles();
   #ifdef VISUAL
      Sleep(1000/4);
   #else
      usleep(1000000/4);
   #endif
   b = clockCycles();
   return (long)(4*(b-a-1401000-450000));
}



//////////////////////////////
//
// Timer::reset -- set the timer to 0.
//

void Timer::reset(void) {
   offset = clockCycles();
}



//////////////////////////////
//
// Timer::setCpuSpeed -- You can also vary the CPU speed here to cause the 
//	getTime function to output different units of time, but the 
//	setTicksPerSecond is a more appropriate place to do such a thing. 
//

void Timer::setCpuSpeed(long aSpeed) {
   if (aSpeed <= 0) {
      cerr << "Error: Cannot set the cpu speed to be negative: " 
           << aSpeed << endl;
      exit(1);
   }
   cpuSpeed = aSpeed;
}



//////////////////////////////
//
// Timer::setPeriod -- sets the period length of the timer.
//    input value cannot be less than 1.0.
//

void Timer::setPeriod(double aPeriod) {
   if (aPeriod < 1.0) {
      cerr << "Error: period too small: " << aPeriod << endl;
      exit(1);
   }
   period = aPeriod;
}



//////////////////////////////
//
// Timer::setTempo -- sets the period length in terms of 
//	beats per minute.
//

void Timer::setTempo(double beatsPerMinute) {
   if (beatsPerMinute < 1.0) {
      cerr << "Error: tempo is too slow: " << beatsPerMinute << endl;
      exit(1);
   }
   period = getTicksPerSecond() * 60.0 / beatsPerMinute;
}



//////////////////////////////
//
// Timer::setTicksPerSecond 
//

void Timer::setTicksPerSecond(long aTickRate) {
   if (aTickRate <= 0) {
      cerr << "Error: Cannot set the tick rate to be negative: " 
           << aTickRate << endl;
      exit(1);
   }
   ticksPerSecond = aTickRate;
}



//////////////////////////////
//
// Timer::start 
//

void Timer::start(void) {
   reset();
}



//////////////////////////////
//
// Timer::sync 
//

void Timer::sync(Timer& aTimer) {
   offset = aTimer.offset;
}



//////////////////////////////
//
// Timer::update -- set the timer start to the next period.
//

void Timer::update(void) {
   if (getTime() >= getPeriod()) {
      offset += (int64bits)(getPeriod() * getFactor());
   }
}



// will do nothing if the periodCount is greater than
// than the expired period count
void Timer::update(int periodCount) {
   if (periodCount > expired()) {
      return;
   } else {
      offset += (int64bits)(getPeriod() * getFactor() * periodCount);
   }
}



///////////////////////////////////////////////////////////////////////////


